use core::fmt::Write;
use embassy_sync::{blocking_mutex::raw::ThreadModeRawMutex, mutex::Mutex};
use heapless::{String, Vec};
use percent_encoding::percent_decode_str;
use pico_website::unwrap;
use ringbuffer::{ConstGenericRingBuffer, RingBuffer};

use crate::socket::{HttpRequestType, HttpResCode};

use super::App;

const MEMORY_SIZE: usize = 16;
const USERNAME_MIN_SIZE: usize = 3;
const USERNAME_SIZE: usize = 16;
const MSG_SIZE: usize = 256;

static MESSAGES: Mutex<ThreadModeRawMutex, Messages> = Mutex::new(Messages::new());

pub struct ChatApp {
    res_buf: Vec<u8, 2048>,
}
impl ChatApp {
    pub fn new() -> Self {
        Self {
            res_buf: Vec::new(),
        }
    }
}
impl App for ChatApp {
    fn socket_name(&self) -> &'static str {
        "chat"
    }
    async fn handle_request<'a>(
        &'a mut self,
        path: &str,
        req_type: HttpRequestType,
        content: &str,
    ) -> (HttpResCode, &'static str, &'a [u8]) {
        match (req_type, path) {
            (HttpRequestType::Get, "/" | "/index" | "/index.html" | "/chat" | "/chat.html") => {
                (HttpResCode::Ok, "html", include_bytes!("./chat.html"))
            }
            (_, path) => {
                let (path, args) = path.split_once('?').unwrap_or((path, ""));
                let mut load = None;
                let mut username = None;
                // let mut msg_content = None;
                for arg in args.split('&').chain(content.split('&')) {
                    match arg.split_once('=') {
                        Some(("load", n)) => {
                            let n: u16 = match n.parse() {
                                Ok(v) => v,
                                Err(_) => return (HttpResCode::BadRequest, "", &[]),
                            };
                            if n > 0 {
                                load = Some(n);
                            }
                        }
                        Some(("username", u)) => {
                            if u.len() < USERNAME_MIN_SIZE || u.len() > USERNAME_SIZE {
                                return (HttpResCode::BadRequest, "", &[]);
                            }
                            username = Some(u);
                        }
                        Some(("msg", m)) => {
                            let mut msg = String::<MSG_SIZE>::new();
                            for c in percent_decode_str(m) {
                                msg.push(c as char);
                            }
                            if m.len() > MSG_SIZE {
                                return (HttpResCode::BadRequest, "", &[]);
                            }
                            // msg_content = Some();
                        }
                        _ => {}
                    }
                }
                if path == "/chat/connect" && username.is_some() {
                    self.res_buf.clear();
                    unwrap(write!(
                        &mut self.res_buf,
                        "<div \
                            class=\"message\" \
                            hx-get=\"/chat/message/abs/0?load={}&username={}\" \
                            hx-target=\"this\" \
                            hx-swap=\"outerHTML\" \
                            hx-trigger=\"load\" \
                        ></div>",
                        MEMORY_SIZE,
                        username.unwrap()
                    ))
                    .await;
                    return (HttpResCode::Ok, "html", &self.res_buf);
                } else if path == "/chat/send" && username.is_some() {
                    return (HttpResCode::Ok, "html", &self.res_buf);
                } else if path.starts_with("/chat/message/abs/") && path.len() > 18 {
                    let msg_id: u16 = match path[18..].parse() {
                        Ok(n) => n,
                        Err(_) => return (HttpResCode::BadRequest, "", &[]),
                    };
                    let msgs = MESSAGES.lock().await;
                    if msg_id >= msgs.next {
                        return (HttpResCode::BadRequest, "", &[]);
                    }
                    if msg_id < msgs.next.saturating_sub(MEMORY_SIZE as u16 + 1) {
                        return (HttpResCode::NoContent, "", &[]);
                    }
                    self.res_buf.clear();
                    unwrap(write!(&mut self.res_buf, "<div class=\"message\"")).await;
                    if let Some(n) = load {
                        unwrap(write!(
                            &mut self.res_buf,
                            " hx-get=\"/chat/message/abs/{}?load={}\" \
                            hx-target=\"this\" \
                            hx-swap=\"afterend\" \
                            hx-trigger=\"load\"",
                            msg_id + 1,
                            n - 1
                        ))
                        .await;
                    }
                    let msg = match msgs.get_abs(msg_id) {
                        Some(msg) => msg,
                        None => return (HttpResCode::NoContent, "", &[]),
                    };
                    unwrap(write!(
                        &mut self.res_buf,
                        "><b>{}</b>: {}</div>",
                        msg.author, msg.content
                    ))
                    .await;
                    return (HttpResCode::Ok, "html", &self.res_buf);
                } else {
                    (HttpResCode::NotFound, "", &[])
                }
            }
        }
    }
}

struct Message {
    author: String<USERNAME_SIZE>,
    content: String<MSG_SIZE>,
}

struct Messages {
    inner: ConstGenericRingBuffer<Message, MEMORY_SIZE>,
    next: u16,
}
impl Messages {
    const fn new() -> Self {
        Self {
            inner: ConstGenericRingBuffer::new(),
            next: 2,
        }
    }
    fn get_abs(&self, id: u16) -> Option<&Message> {
        self.inner
            .get_signed((id as isize) + 1 - (self.next as isize))
    }
}
